import os
from http import HTTPStatus
from typing import TYPE_CHECKING, NamedTuple
from unittest import mock

import pytest
from pytest_mock import MockerFixture

if TYPE_CHECKING:
    from _pytest.config import Config
    from _pytest.config.argparsing import Parser
    from _pytest.nodes import Item

e2e = pytest.mark.e2e


def pytest_addoption(parser: "Parser"):
    parser.addoption("--e2e", action="store_true", default=False, help="run end-to-end tests")


def pytest_collection_modifyitems(config: "Config", items: list["Item"]) -> None:
    if config.getoption("--e2e"):
        required_env_variables = ["E2E_SMTP_EMAIL", "E2E_SMTP_X_API_KEY", "E2E_TARGET_DOMAIN"]
        if missing := [v for v in required_env_variables if not os.getenv(v)]:
            pytest.exit(f"Missing required environment variables for --e2e: {", ".join(missing)}")
        return

    skip_e2e = pytest.mark.skip(reason="use --e2e to run end-to-end tests")
    for item in items:
        if "e2e" in item.keywords:
            item.add_marker(skip_e2e)


@pytest.fixture
def mocked_client(mocker: MockerFixture) -> mock.Mock:
    client: mock.Mock = mocker.Mock()
    client.is_closed = False
    client.base_url = None
    client.headers = {}
    return client


class ExpectedResponseData(NamedTuple):
    code: HTTPStatus
    url: str
    headers: dict[str, str]
    content: bytes


@pytest.fixture
def mocked_urlopen(mocker: MockerFixture) -> ExpectedResponseData:
    data = ExpectedResponseData(HTTPStatus.OK, "https://foo.com/", {"X": "Y"}, b'{"foo": "bar"}')
    mocked_response = mocker.Mock()
    mocked_response.status = data.code
    mocked_response.geturl.return_value = data.url
    mocked_response.headers = data.headers
    mocked_response.read.return_value = data.content
    mocker.patch("boox.client.urlopen").return_value.__enter__.return_value = mocked_response
    return data



"""Example on how to synchronize session token.

While this method is not required at all, especially if you write short scripts,
it makes sense to use it occasionally to make sure that you don't use an expired token.
"""

import datetime
from typing import cast

from boox.core import Boox
from boox.models.enums import BooxUrl
from boox.models.users import SyncTokenResponse

TOKEN = cast(str, ...)  # Just to make this example easier


with Boox(base_url=BooxUrl.EUR, token=TOKEN) as boox:
    response: SyncTokenResponse = boox.users.synchronize_token()

# Although the original response has tokenExpiredAt key, it is aliased for more pythonic use

# In the original response it is an integer, which in fact is a UNIX epoch timestamp.
# For example 1768815112 is an equivalent of GMT: Monday, 19 January 2026 09:31:52.
# For convenience, it is being cast to the datetime.datetime object during data validation.
_: datetime.datetime = response.token_expired_at



import datetime
import re
from contextlib import suppress
from typing import Annotated, Any, ClassVar, Self

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PlainSerializer,
    SecretStr,
    StringConstraints,
    model_validator,
    validate_email,
)
from pydantic_core import PydanticCustomError

from boox.models.base import BaseResponse


def soft_validate_email(value: str) -> bool:
    with suppress(PydanticCustomError):
        return bool(validate_email(value))
    return False


class BaseVerificationModel(BaseModel):
    """The base request body for verification and authentication requests.

    Attributes:
        area_code (str | None): Optional. Required if `mobi` is an e-mail.
        mobi (str): Required. Either mobile number or e-mail.

    """

    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, str_strip_whitespace=True)

    mobi: Annotated[str, StringConstraints(min_length=6)]
    area_code: Annotated[str, StringConstraints(min_length=2, pattern=r"^\+\d+$")] | None = Field(default=None)

    @model_validator(mode="after")
    def check_area_code(self) -> Self:
        is_phone = bool(re.fullmatch(r"\d+", self.mobi))
        is_email = bool(soft_validate_email(self.mobi))

        if not is_phone and not is_email:
            raise ValueError("mobi field must either be an e-mail or a phone number")
        if is_phone and not self.area_code:
            raise ValueError("area_code must be provided if phone method is used")
        if is_email and self.area_code:
            raise ValueError("mobi and area_code are mutually exclusive")
        return self


class SendVerifyCodeRequest(BaseVerificationModel):
    """A request body for POST users/sendVerifyCode.

    Please refer to BaseVerificationModel for more attributes.

    Attributes:
        scene (str): Unknown.
        verify (str): Unknown.

    There are basically 2 important usages:
    - send code to mobile,
    - send code to email.

    Examples:
        >>> SendVerifyCodeRequest(area_code="+48", mobi="600123456")  # mobile
        >>> SendVerifyCodeRequest(mobi="foo@bar.com")  # email
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, str_strip_whitespace=True)

    scene: str = ""
    verify: str = ""


class SendVerifyResponse(BaseResponse[str]):
    """Base response type with data being a str."""


class FetchTokenRequest(BaseVerificationModel):
    """A request body for POST users/signupByPhoneOrEmail.

    Attributes:
        code (str): The verification code received via either mobile or e-mail.

    Note:
        Verification code is 6 digits long, but is passed as a string.
    """

    code: Annotated[str, StringConstraints(pattern=r"^\d{6}$")]


class DataToken(BaseModel):
    token: Annotated[SecretStr, PlainSerializer(lambda v: v.get_secret_value(), return_type=str, when_used="always")]


class FetchTokenResponse(BaseResponse[DataToken]):
    """A response, with token under data key."""


class DataSession(BaseModel):
    channels: tuple[Any, ...]
    cookie_name: str
    expires: datetime.datetime
    session_id: SecretStr


class SyncTokenResponse(BaseResponse[DataSession]):
    """A response, with information about token and session expiration."""

    model_config: ClassVar[ConfigDict] = ConfigDict(serialize_by_alias=True)

    token_expired_at: datetime.datetime = Field(alias="tokenExpiredAt")



from collections.abc import Callable
from functools import wraps
from typing import TYPE_CHECKING, Any, Concatenate, Protocol
from urllib.parse import urljoin

if TYPE_CHECKING:
    from boox.core import Boox


class HasBooxSession(Protocol):
    _session: "Boox"


class TokenMissingError(Exception):
    pass


class Api:
    """An abstract representation of a base API class.

    Although it doesn't inherit from ABC, this class is not meant to be used as a standalone class.
    """

    def __init__(self, session: "Boox"):
        if type(self) is Api:
            raise TypeError("Cannot instantiate abstract class Api directly")
        self._session = session

    def _prepare_url(self, endpoint: str) -> str:
        if self._session.base_url is None:
            msg = f"{type(self._session).__name__}.base_url must be filled"
            raise ValueError(msg)
        return urljoin(self._session.base_url.rstrip("/"), endpoint)

    def _post(self, *, endpoint: str, json: Any | None = None):
        return self._session.client.post(self._prepare_url(endpoint), json=json).raise_for_status()

    def _get(self, *, endpoint: str):
        return self._session.client.get(self._prepare_url(endpoint)).raise_for_status()


def requires_token[T: Api, **P, R](func: Callable[Concatenate[T, P], R]) -> Callable[Concatenate[T, P], R]:
    @wraps(func)
    def wrapper(self: T, *args: P.args, **kwargs: P.kwargs) -> R:
        if not self._session.token:  # pyright: ignore[reportPrivateUsage]
            raise TokenMissingError("Bearer token is required to call this method")
        return func(self, *args, **kwargs)

    return wrapper



from collections.abc import MutableMapping
from typing import Any, Protocol, Self, runtime_checkable


@runtime_checkable
class HttpResponse(Protocol):
    """The minimal requirement for a response to work with the Boox class."""

    def raise_for_status(self) -> Self:
        """A method that either raises an error for non-2xx status code or passes self down."""
        ...

    def json(self, **kwargs: Any) -> Any:
        """A method that returns whatever json.loads would return, typically a dictionary."""
        ...


@runtime_checkable
class HttpClient(Protocol):
    """The minimal requirement for a client to work with the Boox class."""

    headers: Any | MutableMapping[str, str]

    def post(self, url: str, json: Any | None = None, *args: Any, **kwargs: Any) -> HttpResponse:
        """A method to request a `POST` type response."""
        ...

    def get(self, url: str, *args: Any, **kwargs: Any) -> HttpResponse:
        """A method to request a `GET` type response."""
        ...

    def close(self) -> None:
        """A method to close the current or global connection."""
        ...



from boox.models.base import BaseResponse


def test_base_response_str_format():
    response = BaseResponse[None](data=None, message="foo", result_code=123)
    assert str(response) == "<123: foo>"


def test_subclass_response_repr_format():
    class DummyResponse(BaseResponse[None]): ...

    response = DummyResponse(data=None, message="foo", result_code=123)
    assert repr(response) == "DummyResponse(<123: foo>)"



import json as jsonlib
from http import HTTPStatus
from typing import Any, Self
from urllib import error
from urllib.request import Request, urlopen

from pydantic import validate_call


class BaseHTTPError(Exception):
    """Base HTTP exception."""

    def __init__(self, url: str, code: HTTPStatus, headers: dict[str, str], content: bytes) -> None:
        self.url = url
        self.code = code
        self.headers = headers
        self._content = content

    @property
    def content(self) -> str:
        return self._content.decode()

    def __str__(self) -> str:
        return f"HTTP Error {self.code.value}: {self.code.phrase}"

    def __repr__(self) -> str:
        return f"<HTTP Error {self.code.value}: {self.code.phrase!r}>"


class BaseHttpResponse:
    def __init__(self, *, code: HTTPStatus, url: str, headers: dict[str, str], content: bytes) -> None:
        self._code = code
        self._url = url
        self._headers = headers
        self._content = content

    def __repr__(self) -> str:
        return f"<Response [{self._code.value}]>"

    def raise_for_status(self) -> Self:
        if self._code.is_success:
            return self
        raise BaseHTTPError(url=self._url, code=self._code, headers=self._headers, content=self._content)

    def json(self, **kwargs: Any) -> Any:
        return jsonlib.loads(self._content, **kwargs)


class BaseHttpClient:
    def __init__(self) -> None:
        self.headers: dict[str, str] = {}

    def _build_request(self, method: str, url: str, json: Any | None) -> Request:
        data = jsonlib.dumps(json).encode() if json else None
        return Request(url=url, data=data, headers=self.headers, method=method)

    @staticmethod
    def _send(request: Request) -> BaseHttpResponse:
        try:
            with urlopen(request) as r:
                content = r.read()
                status = r.status
                url = r.geturl()
                headers = dict(r.headers)
        except error.HTTPError as e:
            content = e.read()
            status = e.status
            url = e.geturl()
            headers = dict(e.headers)
        return BaseHttpResponse(code=HTTPStatus(status), url=url, headers=headers, content=content)

    @validate_call()
    def post(self, url: str, json: Any | None = None) -> BaseHttpResponse:
        request = self._build_request("POST", url, json)
        return self._send(request)

    @validate_call()
    def get(self, url: str) -> BaseHttpResponse:
        request = self._build_request("GET", url, None)
        return self._send(request)

    @validate_call()
    def delete(self, url: str) -> BaseHttpResponse:
        request = self._build_request("DELETE", url, None)
        return self._send(request)

    def close(self) -> None:
        pass



import datetime
import re
from unittest import mock
from uuid import uuid1

import pytest
from pytest_mock import MockerFixture

from boox.api.core import TokenMissingError
from boox.api.users import UsersApi
from boox.core import Boox
from boox.models.enums import BooxUrl
from boox.models.users import (
    DataSession,
    DataToken,
    FetchTokenRequest,
    FetchTokenResponse,
    SendVerifyCodeRequest,
    SendVerifyResponse,
    SyncTokenResponse,
)
from tests.api.utils import E2EConfig, EmailProvider
from tests.conftest import e2e

# pyright: reportPrivateUsage=false


def test_boox_client_initializes_users_api(mocked_client: mock.Mock):
    boox = Boox(client=mocked_client)
    assert boox.users._session is boox


def test_send_verification_code_calls_post_and_parses_response(mocker: MockerFixture):
    mocked_session = mocker.Mock()
    api = UsersApi(session=mocked_session)

    mocked_response = mocker.Mock()
    mocked_response.json = mocker.Mock(return_value={"data": "ok", "message": "SUCCESS", "result_code": 0})

    api._post = mocker.Mock(return_value=mocked_response)

    payload = SendVerifyCodeRequest.model_validate({"mobi": "foo@bar.com"})
    result = api.send_verification_code(payload=payload)

    assert result == SendVerifyResponse(data="ok", message="SUCCESS", result_code=0)
    api._post.assert_called_once_with(endpoint="/api/1/users/sendVerifyCode", json={"mobi": "foo@bar.com"})


@pytest.mark.parametrize("url", list(BooxUrl))
def test_users_api_send_verification_code_integration(mocker: MockerFixture, mocked_client: mock.Mock, url: BooxUrl):
    mocked_response = mocker.Mock()
    mocked_response.json.return_value = {"data": "ok", "message": "SUCCESS", "result_code": 0}
    mocked_response.raise_for_status.return_value = mocked_response
    mocked_client.post.return_value = mocked_response

    with Boox(client=mocked_client, base_url=url) as boox:
        payload = SendVerifyCodeRequest.model_validate({"mobi": "foo@bar.com"})
        result = boox.users.send_verification_code(payload=payload)

    expected_url = url.value + "/api/1/users/sendVerifyCode"
    expected_json = payload.model_dump(exclude_unset=True)
    mocked_client.post.assert_called_once_with(expected_url, json=expected_json)
    mocked_response.json.assert_called_once()
    assert isinstance(result, SendVerifyResponse)
    assert result.data == "ok"


def test_fetch_session_token_calls_post_and_parses_response(mocker: MockerFixture):
    token = str(uuid1())
    mocked_session = mocker.Mock()
    api = UsersApi(session=mocked_session)

    mocked_response = mocker.Mock()
    mocked_response.json = mocker.Mock(return_value={"data": {"token": token}, "message": "SUCCESS", "result_code": 0})

    api._post = mocker.Mock(return_value=mocked_response)

    payload = FetchTokenRequest.model_validate({"mobi": "foo@bar.com", "code": "123456"})
    result = api.fetch_session_token(payload=payload)

    data = FetchTokenResponse.model_validate({"data": {"token": token}, "message": "SUCCESS", "result_code": 0})
    assert result == data
    api._post.assert_called_once_with(
        endpoint="/api/1/users/signupByPhoneOrEmail", json={"mobi": "foo@bar.com", "code": "123456"}
    )


@pytest.mark.parametrize("url", list(BooxUrl))
def test_users_api_fetch_session_token_integration(mocker: MockerFixture, mocked_client: mock.Mock, url: BooxUrl):
    token = str(uuid1())
    mocked_response = mocker.Mock()
    mocked_response.json.return_value = {"data": {"token": token}, "message": "SUCCESS", "result_code": 0}
    mocked_response.raise_for_status.return_value = mocked_response
    mocked_client.post.return_value = mocked_response

    with Boox(client=mocked_client, base_url=url) as boox:
        payload = FetchTokenRequest.model_validate({"mobi": "foo@bar.com", "code": "123456"})
        result = boox.users.fetch_session_token(payload=payload)

    expected_url = url.value + "/api/1/users/signupByPhoneOrEmail"
    expected_json = payload.model_dump(exclude_unset=True)
    mocked_client.post.assert_called_once_with(expected_url, json=expected_json)
    mocked_response.json.assert_called_once()
    assert isinstance(result, FetchTokenResponse)
    data = DataToken.model_validate({"token": token})
    assert result.data == data


@pytest.mark.parametrize("url", list(BooxUrl))
def test_users_api_sync_token_integration(mocker: MockerFixture, mocked_client: mock.Mock, url: BooxUrl):
    token = str(uuid1())
    mocked_response = mocker.Mock()
    data = {
        "channels": (),
        "cookie_name": "foo",
        "expires": datetime.datetime.now(tz=datetime.UTC),
        "session_id": "xyz123",
    }
    mocked_response.json = mocker.Mock(
        return_value={"data": data, "message": "SUCCESS", "result_code": 0, "tokenExpiredAt": 1}
    )
    mocked_response.raise_for_status.return_value = mocked_response
    mocked_client.get.return_value = mocked_response

    with Boox(client=mocked_client, base_url=url, token=token) as boox:
        result = boox.users.synchronize_token()

    expected_url = url.value + "/api/1/users/syncToken"
    mocked_client.get.assert_called_once_with(expected_url)
    mocked_response.json.assert_called_once()
    assert isinstance(result, SyncTokenResponse)
    assert result.data == DataSession.model_validate(data)


@pytest.mark.parametrize("url", list(BooxUrl))
def test_2(mocked_client: mock.Mock, url: BooxUrl):
    with (
        Boox(client=mocked_client, base_url=url) as boox,
        pytest.raises(TokenMissingError, match="Bearer token is required to call this method"),
    ):
        boox.users.synchronize_token()


@e2e
@pytest.mark.order(0)
def test_send_verification_code_e2e(config: E2EConfig, email: EmailProvider):
    payload = SendVerifyCodeRequest.model_validate({"mobi": config.email_address})

    with Boox(base_url=config.domain) as boox:
        response = boox.users.send_verification_code(payload=payload)

    assert response.data == "ok"
    message = email.get_newest_message()
    match = re.compile(r"^The code is (?P<code>\d{6}) for account verification from BOOX\.").match(message)
    assert match, "Did not match the received email"
    config.verification_code = match.group("code")


@e2e
@pytest.mark.order(1)
def test_fetch_session_token_e2e(config: E2EConfig):
    if not config.verification_code:
        pytest.skip("Verification code was either not obtained or not set")
    payload = FetchTokenRequest.model_validate({"mobi": config.email_address, "code": config.verification_code})

    with Boox(base_url=config.domain) as boox:
        response = boox.users.fetch_session_token(payload=payload)

    assert response.data.token
    config.token = response.data.token


@e2e
def test_synchronize_token_e2e(config: E2EConfig):
    if not config.token:
        pytest.skip("Token was either not obtainer or not set")

    with Boox(base_url=config.domain, token=config.token) as boox:
        response = boox.users.synchronize_token()

    assert response.token_expired_at



from pydantic import validate_call

from boox.api.core import Api, requires_token
from boox.models.users import (
    FetchTokenRequest,
    FetchTokenResponse,
    SendVerifyCodeRequest,
    SendVerifyResponse,
    SyncTokenResponse,
)


class UsersApi(Api):
    """API wrappers for `users` endpoint family.

    Note that since Boox class already has UsersApi in its context,
    it is not recommended to use UsersApi as a standalone object.
    """

    @validate_call()
    def send_verification_code(self, *, payload: SendVerifyCodeRequest) -> SendVerifyResponse:
        """Initial call to get the verification code.

        Depending on the payload, it will either send the verification code on provided e-mail or phone number.
        For phone numbers, works internationally, depending on the provided area code.

        The verification code is valid for 5 minutes.
        The official BOOXDrop service has a 1 minute countdown before you can resend the code (for a particular method).

        This method is used **before** authentication. Luckily, it is not necessary to use it every single time,
        because the tokens received after verification expire every 20 days.

        Args:
            payload (SendVerifyCodeRequest): The validated payload to be sent in order to receive the verification code.

        Returns:
            SendVerifyResponse: The validated, generic response that is always received from the server.
        """
        response = self._post(endpoint="/api/1/users/sendVerifyCode", json=payload.model_dump(exclude_unset=True))
        return SendVerifyResponse.model_validate(response.json())

    @validate_call()
    def fetch_session_token(self, *, payload: FetchTokenRequest) -> FetchTokenResponse:
        """A call to sign in using the obtained verification code.

        Once a verification code is received, it can be used for this request in a payload, along the verification method used in
        send_verification_code method above.

        This method is the vertification. It allows you to receive the session token used in many other requests of all types.

        Args:
            payload (FetchTokenRequest): The validated payload to be sent in order to fetch the session token.

        Returns:
            FetchTokenResponse: The validated response containing a session token.
        """
        response = self._post(endpoint="/api/1/users/signupByPhoneOrEmail", json=payload.model_dump(exclude_unset=True))
        return FetchTokenResponse.model_validate(response.json())

    @requires_token
    def synchronize_token(self) -> SyncTokenResponse:
        """A call to check token authenticity and validity.

        A typical scenario for this route is to use it before any action to prevent a server error.

        This call **requires** the token to be passed as an Authorization header, e.g.:
            >>> {"Authorization": "Bearer xyz123abc"}
        That is also the reason why this call pre-validates the client header.

        Returns:
            SyncTokenResponse: The validated response containing information about token expiry date, and session metadata.
        """
        response = self._get(endpoint="/api/1/users/syncToken")
        return SyncTokenResponse.model_validate(response.json())



import os
import re
import time
from collections.abc import Callable
from functools import wraps
from typing import Self
from urllib.parse import urljoin

from pydantic import SecretStr

from boox.client import BaseHttpClient, BaseHTTPError
from boox.models.enums import BooxUrl


def with_retry[**P, T](
    *, retries: int, delay: int, exceptions: tuple[type[BaseException], ...]
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            result: T | None = None
            for attempt in range(retries):
                try:
                    result = func(*args, **kwargs)
                    break
                except exceptions:
                    if attempt == retries - 1:
                        raise
                    time.sleep(delay)
            if not result:
                raise ValueError("No value under result")
            return result

        return wrapper

    return decorator


class E2EConfig:
    _instance = None
    _initialized: bool = False

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        if self._initialized:
            return
        self.domain: BooxUrl = BooxUrl(os.environ["E2E_TARGET_DOMAIN"])
        self.email_address: str = os.environ["E2E_SMTP_EMAIL"]
        self.x_api_key: str = os.environ["E2E_SMTP_X_API_KEY"]
        self._verification_code: str | None = None
        self._token: SecretStr = SecretStr("")
        self._initialized = True

    @property
    def verification_code(self) -> str | None:
        return self._verification_code

    @verification_code.setter
    def verification_code(self, value: str):
        match = re.compile(r"\d{6}").fullmatch(value)
        if not match:
            raise ValueError("Verification code must contain 6 digits")
        self._verification_code = match.group()

    @property
    def token(self) -> SecretStr:
        return self._token

    @token.setter
    def token(self, value: SecretStr):
        self._token = value


class EmailProvider:
    def __init__(self, config: E2EConfig):
        self.config = config
        self.address = self.config.email_address
        self.client = BaseHttpClient()
        self.client.headers.update({
            "X-API-KEY": self.config.x_api_key,
            "Accept": "application/ld+json",
        })
        self.base_url = "https://api.smtp.dev"
        self.messages_url: str | None = None
        self.newest_message_url: str | None = None

    def _prepare_url(self, route: str) -> str:
        return urljoin(self.base_url, route)

    @with_retry(retries=5, delay=1, exceptions=(BaseHTTPError, IndexError, KeyError))
    def _get_mailbox_url(self) -> str:
        response = self.client.get(self._prepare_url("/accounts"))
        data = response.raise_for_status().json()

        member = data["member"][0]
        mailboxes: list[dict[str, str]] = member["mailboxes"]
        inbox = next(filter(lambda m: m["path"] == "INBOX", mailboxes))
        return inbox["@id"]

    @with_retry(retries=5, delay=1, exceptions=(BaseHTTPError, IndexError, KeyError))
    def _get_message_url(self) -> str:
        if not self.messages_url:
            raise ValueError("No INBOX messages url obtained yet")

        response = self.client.get(self._prepare_url(self.messages_url))
        data = response.raise_for_status().json()

        member: dict[str, str] = data["member"][0]
        return f"{self.messages_url}/{member["id"]}"

    @with_retry(retries=10, delay=1, exceptions=(BaseHTTPError, IndexError, KeyError))
    def _get_newest_message(self) -> str:
        if not self.newest_message_url:
            raise ValueError("No newest message url obtained yet")

        response = self.client.get(self._prepare_url(self.newest_message_url))
        data = response.raise_for_status().json()

        return data["text"]

    def get_newest_message(self) -> str:
        mailbox_url = self._get_mailbox_url()
        self.messages_url = f"{mailbox_url}/messages"
        self.newest_message_url = self._get_message_url()
        return self._get_newest_message()

    def cleanup_inbox(self):
        if not self.messages_url:
            raise ValueError("No INBOX messages url obtained yet")

        response = self.client.get(self._prepare_url(self.messages_url))
        data = response.raise_for_status().json()

        messages: list[dict[str, str]] = data["member"]
        message_ids = [m["id"] for m in messages]
        for message_id in message_ids:
            self.client.delete(self._prepare_url(f"{self.messages_url}/{message_id}")).raise_for_status()



from pydantic import BaseModel


class BaseResponse[T](BaseModel):
    """General server response.

    Attributes:
        data (T | None): Arbitrary response data.
        message (str): The response message.
        result_code (int): Internal result code.
    """

    data: T
    message: str
    result_code: int

    def __str__(self) -> str:
        return f"<{self.result_code}: {self.message}>"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self!s})"



from http import HTTPStatus
from unittest import mock
from uuid import uuid1

import pytest
from pytest_mock import MockerFixture

from boox.api.core import Api, TokenMissingError, requires_token
from boox.client import BaseHTTPError
from boox.core import Boox
from boox.models.enums import BooxUrl

# pyright: reportPrivateUsage=false


class DummyApi(Api):
    @requires_token
    def dummy_method(self):
        return self


def test_api_cannot_be_instantiated_directly(mocker: MockerFixture):
    with pytest.raises(TypeError, match=r"Cannot instantiate abstract class Api directly"):
        Api(session=mocker.Mock())


def test_prepare_url_raises_without_base_url(mocker: MockerFixture):
    mocked_session = mocker.Mock()
    mocked_session.base_url = None
    api = DummyApi(session=mocked_session)
    with pytest.raises(ValueError, match=r"base_url must be filled"):
        api._prepare_url("/endpoint")


@pytest.mark.parametrize("url", list(BooxUrl))
def test_prepare_url_joins_base_and_endpoint_without_leading_slash(mocker: MockerFixture, url: BooxUrl):
    mocked_session = mocker.Mock()
    mocked_session.base_url = url
    api = DummyApi(session=mocked_session)

    endpoint = "endpoint"
    assert api._prepare_url(endpoint) == f"{url.value}/{endpoint}"


@pytest.mark.parametrize("url", list(BooxUrl))
def test_prepare_url_joins_base_and_endpoint_with_leading_slash(mocker: MockerFixture, url: BooxUrl):
    mocked_session = mocker.Mock()
    mocked_session.base_url = url
    api = DummyApi(session=mocked_session)
    endpoint = "/endpoint"
    assert api._prepare_url(endpoint) == f"{url.value}{endpoint}"


@pytest.mark.parametrize("url", list(BooxUrl))
def test_prepare_url_strips_trailing_slash(mocker: MockerFixture, url: BooxUrl):
    mocked_session = mocker.Mock()
    mocked_session.base_url = url + "/"
    api = DummyApi(session=mocked_session)
    endpoint = "/endpoint"
    assert api._prepare_url(endpoint) == f"{url.value}{endpoint}"


def test_post_calls_client_and_checks_status(mocker: MockerFixture):
    mocked_response = mocker.Mock()
    mocked_response.raise_for_status = mocker.Mock(return_value=mocked_response)

    mocked_client = mocker.Mock()
    mocked_client.post = mocker.Mock(return_value=mocked_response)

    mocked_session = mocker.Mock()
    mocked_session.client = mocked_client

    api = DummyApi(session=mocked_session)
    api._prepare_url = mocker.patch.object(api, "_prepare_url", return_value="https://foo.com/endpoint")

    result = api._post(endpoint="/endpoint", json={"foo": "bar"})

    api._prepare_url.assert_called_once_with("/endpoint")
    mocked_client.post.assert_called_once_with("https://foo.com/endpoint", json={"foo": "bar"})
    mocked_response.raise_for_status.assert_called_once()
    assert result == mocked_response


def test_post_raises_on_http_error(mocker: MockerFixture):
    url = "https://foo.com/endpoint"
    mocked_response = mocker.Mock()
    mocked_response.raise_for_status = mocker.Mock(side_effect=BaseHTTPError(url, HTTPStatus.BAD_REQUEST, {}, b''))

    mocked_client = mocker.Mock()
    mocked_client.post = mocker.Mock(return_value=mocked_response)

    mocked_session = mocker.Mock()
    mocked_session.client = mocked_client

    api = DummyApi(session=mocked_session)
    api._prepare_url = mocker.patch.object(api, "_prepare_url", return_value=url)

    with pytest.raises(BaseHTTPError):
        api._post(endpoint="/endpoint")


def test_method_raises_error_for_missing_token(mocked_client: mock.Mock):
    mocked_session = Boox(client=mocked_client)

    api = DummyApi(session=mocked_session)

    with pytest.raises(TokenMissingError, match="Bearer token is required to call this method"):
        api.dummy_method()


def test_method_succeeds_with_token(mocked_client: mock.Mock):
    mocked_session = Boox(client=mocked_client)
    mocked_session.token = str(uuid1())

    api = DummyApi(session=mocked_session)

    assert api.dummy_method()


def test_get_calls_client_and_checks_status(mocker: MockerFixture):
    mocked_response = mocker.Mock()
    mocked_response.raise_for_status = mocker.Mock(return_value=mocked_response)

    mocked_client = mocker.Mock()
    mocked_client.get = mocker.Mock(return_value=mocked_response)

    mocked_session = mocker.Mock()
    mocked_session.client = mocked_client

    api = DummyApi(session=mocked_session)
    api._prepare_url = mocker.patch.object(api, "_prepare_url", return_value="https://foo.com/endpoint")

    result = api._get(endpoint="/endpoint")

    api._prepare_url.assert_called_once_with("/endpoint")
    mocked_client.get.assert_called_once_with("https://foo.com/endpoint")
    mocked_response.raise_for_status.assert_called_once()
    assert result == mocked_response
